

// Feather9x_TX  node Id 0x68 to connect to server 0xb6
/**
   Select Feather MO as the board to compile this sketch
*/
/*
   (c)2014-2018 Forward Computing and Control Pty. Ltd.
   NSW Australia, www.forward.com.au
   This code is not warranted to be fit for any purpose. You may only use it at your own risk.
   This code may be freely used for both private and commercial use
   Provide this copyright is maintained.
*/

/*
   Outline of logic
   Normally one of the pushbuttons is pressed to start, after a couple of seconds the setup() runs (why does Adafruit delay the startup?)
   The pushbutton need to be pressed for about 1sec so that that plus the 1.2sec RC time constant keeps the battery attached long enough for the setup() to run
   The setup() turns on the output to keep the battery connected and then reads the switches.  The DebouncedSwitch constructor initializes the class with the switch position on startup.

   If neither pushbutton has been pressed then it is assumed the USB charging cable has been plugged in an the charging Flash sequence is started and the connection to the battery kept open.
   i.e. never times out.
   Otherwise the battery voltage is check to see if it needs charging (<3.65V). If so the needsCharging flag is set.

   If in charging mode, pressing any pushbutton will exit charging mode and disconnect the battery.
   Note: Since charging mode does not timeout, removing the USB supply leaves the using running on the battery.
   A pushbutton needs to be pressed to turn off the charging mode, AFTER the USB supply is removed, to disconnect the battery.  After that to reenter charging mode
   the USB supply needs to be removed and re-connected.

   If the USB supply is still connected then the micro will continue to run and after 6 sec the battery will be reconnected and charging started again.

   If charging, first pushbutton stops charging, then the second press, within 6 sec, will send that command.


*/
//#define DEBUG
#include "Flasher.h"
// expects  initLeds(); led_On(); led_Off(); to be defined in this sketch
#include "Buzzer.h"
// expects initBuzzer(); buzzer_On(); buzzer_Off(); to be defined in this sketch
#include "DebouncedSwitch.h"
#include <pfodDelay.h>

pfodDelay offDelay;
const unsigned long MIN_OFF_TIMER_INTERVAL = 3000; // 3 sec keep led on for this min time
const unsigned long ERROR_OFF_TIMER_INTERVAL = 5000; // 5 sec error flash if failed

pfodDelay restartChargingDelay;
const unsigned long RESTART_CHARGING_INTERVAL = 20000; // 20 sec, is uC still running after turning off the battery then still USB powered, turn battery on again for charging

pfodDelay turnOffChargingDelay;
const unsigned long TURN_OFF_CHARGING_INTERVAL = 2000; // 2 sec before turn led off if USB still connected

#ifdef DEBUG
pfodDelay debugTimer;
const unsigned long DEBUG_TIMER_INTERVAL = 1000; // 1 sec print timer states at this interval
#endif

const int openSwPin = 6;
const int closeSwPin = 5;
DebouncedSwitch openSw(openSwPin); // monitor a switch on input D6
DebouncedSwitch closedSw(closeSwPin); // monitor a switch on input D5

bool openPressed = false;
bool closedPressed = false;
bool charging = false; // if not sw pressed on power up
bool needsCharging = false;

const int buzzerPin = A5;
const int externalLedPin = 12;
const int ledPin = 13;
const int batteryPowerPin = 11;

// 4.19V == 640 ==> 3.65 == 557 count
const int batteryLowVolts = 557; // 3.65V;
int batteryVolts = 0;
const int ADC_PIN = A1; // PIN for adc if you fix the PowerOffCircuitBoard_R1 by cuting track under R5 and cut track from Fet pin to R5 and link Fet pin to BAT
//const int ADC_PIN = A7 // if you instead if omitting R5 and R6 from PowerOffCircuitBoard_R1 connect ADC lead to D9

void printDebugTimers() {
#ifdef DEBUG
  Serial.print(millis()); Serial.print("mS");
  Serial.print(" offDelay:"); Serial.print(offDelay.remaining());
  Serial.print(" restartChargingDelay:"); Serial.print(restartChargingDelay.remaining());
  Serial.print(" turnOffChargingDelay:"); Serial.print(turnOffChargingDelay.remaining());
  Serial.println();
  printRadioDebugTimers();
#endif
}

void debugTimers() {
#ifdef DEBUG
  if (debugTimer.justFinished()) {
    debugTimer.repeat();
    printDebugTimers();
  }
#endif
}

void buzzer_On() {
  digitalWrite(buzzerPin, HIGH);
}
void buzzer_Off() {
  digitalWrite(buzzerPin, LOW);
}

void led_On() {
  digitalWrite(ledPin, HIGH);
  digitalWrite(externalLedPin, HIGH);
}

void led_Off() {
  digitalWrite(ledPin, LOW);
  digitalWrite(externalLedPin, LOW);
}

void initLed() {
  pinMode(ledPin, OUTPUT);
  pinMode(externalLedPin, OUTPUT);
  led_Off();
}

void initBuzzer() {
  pinMode(buzzerPin, OUTPUT);
  buzzer_Off();
}


void turnBatteryOn() {
  pinMode(batteryPowerPin, OUTPUT);
  digitalWrite(batteryPowerPin, HIGH);
}

void turnBatteryOff() {
  setBuzzer(BUZZER_OFF);
  if (charging) { // do not turn off
    setFlash(CHARGING_FLASH);
#ifdef DEBUG
    Serial.println("Charging so don't turn off");
#endif
  } else if (needsCharging) {
    setBuzzer(BUZZER_NEEDS_CHARGING);
    setFlash(CHARGING_FLASH);
    needsCharging = false;
    disableTimers();
    offDelay.start(MIN_OFF_TIMER_INTERVAL);
  } else {
    led_Off();
    disableTimers();
    restartChargingDelay.start(RESTART_CHARGING_INTERVAL);
#ifdef DEBUG
    Serial.println("Turn OFF");
#endif
    digitalWrite(batteryPowerPin, LOW);
  }
}


void disableTimers() {
#ifdef DEBUG
  Serial.println("disableTimers()");
#endif
  offDelay.stop();
  restartChargingDelay.stop(); // disable
  turnOffChargingDelay.stop();
  disableRadioTimers();
}

void checkOffTime() {
  if (offDelay.justFinished()) {
    turnBatteryOff();
  }
}

void checkRestartCharging() {
  if (restartChargingDelay.justFinished()) {
    turnBatteryOn(); // restart charging
    charging = true;
    setFlash(CHARGING_FLASH);
#ifdef DEBUG
    Serial.println("restartCharging");
#endif
  }
}

void checkTurnOffCharging() {
  if (turnOffChargingDelay.justFinished()) {
    turnBatteryOff();
    setFlash(FLASH_OFF);
#ifdef DEBUG
    Serial.println("turnOffCharging()");
#endif
  }
}

void openRequested() {
#ifdef DEBUG
  Serial.println("openRequested()");
#endif
  if (charging) {
    charging = false;
#ifdef DEBUG
    Serial.println("   charging set to false");
#endif
    disableTimers();
    turnOffChargingDelay.start(TURN_OFF_CHARGING_INTERVAL);
    turnBatteryOff();
    return; // ignore first time
  }
  sendOpenRequested(); // disables timers and sets offDelay for 6 sec
#ifdef DEBUG
  Serial.println("   setBuzzer(BUZZER_OPEN)");
#endif
  setBuzzer(BUZZER_OPEN);
}

void closeRequested() {
#ifdef DEBUG
  Serial.println("closeRequested()");
#endif
  if (charging) {
    charging = false;
#ifdef DEBUG
    Serial.println("    charging set to false");
#endif
    disableTimers();
    turnOffChargingDelay.start(TURN_OFF_CHARGING_INTERVAL);
    turnBatteryOff();
    return; // ignore first time
  }
  sendCloseRequested(); // disables timers and sets offDelay for 6 sec
#ifdef DEBUG
  Serial.println("   setBuzzer(BUZZER_CLOSED)");
#endif
  setBuzzer(BUZZER_CLOSED);
}


void setup() {
  batteryVolts = analogRead(ADC_PIN);
#ifdef DEBUG
  Serial.begin(115200);
  for (int i = 10; i > 0; i--) {
    Serial.print(i); Serial.print(' ');
    delay(500);
  }
  Serial.println();
#endif

  turnBatteryOn();

#ifdef DEBUG
  Serial.print("ADC:"); Serial.println(batteryVolts);
#endif

  // put your setup code here, to run once:
  flashInit();
  buzzerInit();
  radioSetup();

  if (closedSw.isDown()) {
    closeRequested();
  } else if (openSw.isDown()) {
    openRequested();
  } else {
    charging = true;
  }

  if (charging) {
    setFlash(CHARGING_FLASH);
    setMsgStateCharging();
  } else {
    // check for battery low here and turn on flasher
    if (batteryVolts <= batteryLowVolts) {
#ifdef DEBUG
      Serial.print("Battery Low ");
      Serial.println(batteryVolts);
#endif
      needsCharging = true;
    }
  }

#ifdef DEBUG
  debugTimer.start(DEBUG_TIMER_INTERVAL);
#endif

}

void loop() {
  debugTimers(); // empty call in not #define DEBUG
  checkBuzzer();
  checkOffTime();
  checkLedFlasher();
  checkRestartCharging();
  checkTurnOffCharging();
  // check the switches
  openSw.update();
  closedSw.update();
  if (closedSw.isChanged() && closedSw.isDown()) {
    closeRequested();
  } else if (openSw.isChanged() && openSw.isDown()) {
    openRequested();
  }
  checkRadio();
}
