#include <SPI.h>
#include <RH_RF95.h>
// download the libraries from http://www.forward.com.au/pfod/pfodParserLibraries/index.html
// pfodParser.zip V3.26+ contains pfodParser, pfodSecurity, pfodBLEBufferedSerial, pfodSMS and pfodRadio
#include <pfodRadio.h>
#include <pfodSecurityClient.h>
// pfodSecurityClient.h includes pfodDelay.h

//#define DEBUG_RADIO

// use last byte of LoRa address as nodeID  (need to check it is unique here)
const uint8_t NODE_ID = 0x68;  // address of this client
const uint8_t SERVER_NODE_ID = 0xb6; // address of pfodServer to connect to

// Change to 434.0 or other frequency, must match the RX's, pfodServer's, freq!  AND must match the range allowed in your country.
#define RF95_FREQ 915.0
// == 81.9mm length wire for 1/4 wave length antenna, see https://www.easycalculation.com/physics/electromagnetism/antenna-wavelength.php

// add your pfod Password here for 128bit security
// eg "b0Ux9akSiwKkwCtcnjTnpWp" but generate your own key, "" means no pfod password
#define pfodSecurityCode "WuxwKgpcEx((UXEqddHstqp"
// see http://www.forward.com.au/pfod/ArduinoWiFi_simple_pfodDevice/index.html for more information and an example
// and QR image key generator.

char *REQUEST_IMAGE_UPDATE_MSG  = "{V3:z`0}";
char *CLOSE_CONNECTION_MSG  = "{!}";
char *DOOR_CLOSED = "Closed"; //
char *DOOR_OPEN = "<g><b>Open";  //
char *DOOR_CLOSING = "Closing"; //
char *DOOR_OPENING = "Opening";  //
char *DOOR_STOPPED = "Stopped"; //

char *CLOSE_DOOR = "{V3:A~C`0`0`0}";  // cmd 'C'
char *OPEN_DOOR = "{V3:A~O`0`0`0}";  //  cmd 'O'

// Radio sends are broken up into blocks of <255 bytes, with measured tx time of 460mS per 255 byte block
// so sends from Server (max 1023 bytes) can take upto 2secs to be received at the client
// Each radio msg is acked. The default ack timeout is 500mS, i.e. random timeout in the range 500mS to 1000mS  (use setAckTimeout(mS) to change this)
// This time allows for another, interfering, radio to finish its attempted transmission
// After 5 retries the link is marked as failed. (use setNoOfRetries() to change this number)
// Most pfod commands are very short, <20 bytes (max 254 bytes), but responses from the pfodServer (pfodDevice) can be upto 1023 bytes.

//==================================== pfodRHDriver ================
// The class to interface with the low lever radio driver.  Must extend from pfodRadioDrive interface
// class implementation is at the bottom of this file.
class pfodRHDriver : public pfodRadioDriver {
  public:
    pfodRHDriver(RHGenericDriver* _driver);  // RHGenericDriver is the low lever RadioHead class used to talk to the radio module
    bool init();  int16_t lastRssi();   int getMode();   uint8_t getMaxMessageLength();   void setThisAddress(uint8_t addr);
    bool receive(uint8_t* buf, uint8_t* len);   bool send(const uint8_t* data, uint8_t len);   uint8_t headerTo();
    uint8_t headerFrom();   uint8_t headerId();   uint8_t headerFlags(); void setHeaderTo(uint8_t to);
    void setHeaderFrom(uint8_t from);   void setHeaderId(uint8_t id);  void setHeaderFlags(uint8_t flags);
  private:
    RHGenericDriver* driver;
};
//============= end of pfodRHDriver header

// defines for RadioHead driver for Feather M0
#define RFM95_CS 8
#define RFM95_RST 4
#define RFM95_INT 3

// Singleton instance of the low level radio driver
RH_RF95 driver(RFM95_CS, RFM95_INT);

pfodRHDriver pfodDriver(&driver); // create the pfodRadioDriver class
pfodRadio radio(&pfodDriver, NODE_ID); // the high level pfod radio driver class
pfodSecurityClient pfodClient;  // the pfodClient with optional 128bit security

Stream* rawDataStream = pfodClient.getRawDataStream();  // the stream where any RawData received is delivered

enum DoorStateEnum {CLOSED, OPEN, OPENING, CLOSING, STOPPED, _UNKNOWN_ };
DoorStateEnum doorState = _UNKNOWN_;
char requestedDoorState = ' '; // none;
char lastRequestedDoorState = requestedDoorState;

enum msgStateEnum {CONNECTED, SENT_INQUIRY, SENT_CMD, RETRY_WAIT, CONNECTION_FINISHED, CONNECTION_ERROR, CONNECTION_SEND_CLOSE, CHARGING};
msgStateEnum msgState = CHARGING; // processMsgs not called unless connected


pfodDelay msgTimer;
const unsigned long MSG_TIMEOUT = 4000; // 4 sec

pfodDelay travelTimer; //
const unsigned long TRAVEL_TIMER_INTERVAL = 25000; // 25 sec for door to open/close allow for change of direction

pfodDelay retryTimer; //
const unsigned long RETRY_TIMER_INTERVAL = 1000; // 1 between retries

unsigned int retryCounter = 0;
unsigned int MAX_RETRIES = 5; // about 15sec to exhaust retries + 20 flashing

const size_t MAX_RETURN_MSG_LEN = 255;
char returnMsg[MAX_RETURN_MSG_LEN + 1];
uint8_t parseReturn(uint8_t c);
void parseReturnInit();
DoorStateEnum checkDoorState();

void setMsgStateCharging() {
  if (msgState == CONNECTION_FINISHED) {
    msgState = CHARGING;
  }
}

void manualResetRadio() {
  // manual radio reset
  digitalWrite(RFM95_RST, LOW);
  delay(20);
  digitalWrite(RFM95_RST, HIGH);
  delay(100); // wait 0.1 sec
}

void sendCloseRequested() {
  requestedDoorState = 'C'; // closed
  lastRequestedDoorState = requestedDoorState;
  disableTimers();
  offDelay.start(MSG_TIMEOUT + 1000); // stay on for at least msg timeout
  retryCounter = 0;
  if ((msgState == CONNECTION_FINISHED) || (msgState == CHARGING))  {
    msgState = CONNECTED;
    setFlash(NO_FLASH);
  }
#ifdef DEBUG_RADIO
  Serial.print("sendCloseRequested() "); Serial.println(debugMsgState());
#endif
}

void sendOpenRequested() {
  requestedDoorState = 'O'; // open
  lastRequestedDoorState = requestedDoorState;
  disableTimers();
  offDelay.start(MSG_TIMEOUT + 1000); // stay on for at least msg timeout
  retryCounter = 0;
  if ((msgState == CONNECTION_FINISHED) || (msgState == CHARGING))  {
    msgState = CONNECTED;
    setFlash(NO_FLASH);
  }
#ifdef DEBUG_RADIO
  Serial.print("sendOpenRequested() "); Serial.println(debugMsgState());
#endif
}

bool isConnectionFinished() {
  return (msgState == CONNECTION_FINISHED);
}

void printRadioDebugTimers() {
#ifdef DEBUG_RADIO
  Serial.print(" travelTimer:"); Serial.print(travelTimer.remaining());
  Serial.print(" retryTimer:"); Serial.print(retryTimer.remaining());
  Serial.print(" msgTimer:"); Serial.print(msgTimer.remaining());
  Serial.print(" retryCounter:"); Serial.print(retryCounter);
  Serial.print(" msgState:"); Serial.print(debugMsgState());
  Serial.println();
#endif
}

void radioSetup() {
#ifdef DEBUG_RADIO
  Serial.println("Feather M0 LoRa pfodClient");
  //radio.setDebugStream(&Serial);  // uncomment this to get debug statements from pfodRadio.  The defines in pfodRadio.cpp determine what is output.
  //pfodClient.setDebugStream(&Serial); // uncomment this to get debug statements from pfodSecurityClient. The defines in pfodSecurityClient.cpp determine what is output.
#endif

  driver.setPromiscuous(false); // only accept our address from low level driver
  // low level drive reset pin setup
  pinMode(RFM95_RST, OUTPUT);
  digitalWrite(RFM95_RST, HIGH);
  bool radioInitialized = false;
  bool freqChanged = false;
  bool configSet = false;
  while ((!radioInitialized) || (!freqChanged) || (!configSet)) {  // loop here until radio inits
    manualResetRadio(); // delays 0.1sec after
    radioInitialized = radio.init();
    if (!radioInitialized) {
#ifdef DEBUG_RADIO
      Serial.println("LoRa radio init failed");
#endif
      delay(1000); // pause and then try again
      continue;
    }
    // setup LoRa radio
    // Defaults after init are 434.0MHz, modulation Bw125Cr45Sf128, +13dbM
    freqChanged = driver.setFrequency(RF95_FREQ);
    if (!freqChanged) {
#ifdef DEBUG_RADIO
      Serial.println("LoRa radio freq set failed");
#endif
      delay(1000); // pause and then try again
      continue;
    }
    configSet = driver.setModemConfig(RH_RF95::Bw125Cr45Sf128); // default
    if (!configSet) {
#ifdef DEBUG_RADIO
      Serial.println("LoRa radio modem config set failed");
#endif
      delay(1000); // pause and then try again
      continue;
    }
  }
#ifdef DEBUG_RADIO
  Serial.println("LoRa radio init OK!");
  Serial.println();
  Serial.println("You can now send pfod cmds from Arduino IDE Serial Monitor");
  Serial.println("  Responses, and any rawData, will be displayed in the Serial Monitor");
  Serial.println("Send ");
  Serial.println("{!}");
  Serial.println(" to close the connection. The connection will be automatically re-established when the next cmd is entered.");
  Serial.println();
#endif

  driver.setTxPower(23, false);  // set max Tx power
  radio.connectTo( SERVER_NODE_ID); // set the server address to connect to
  //radio.setAckTimeout(500); // this is the mS default timeout for acks to each radio msg. Usually no need to change this.
  pfodClient.setResponseTimeout(5000);  // in mS.  Expect { ... } response from pfodServer within 5secs, else connection lost
  //pfodClient.setKeepAliveInterval(5000); // no keep alives
  // default is no keepalives
  // send cmd {!} to close the connection
  // connect will be automatically established again on next command sent

  pfodClient.connect(&radio, F(pfodSecurityCode)); // setup connection. Nothing sent until there is some input

}

bool isConnectionError() {
  return (msgState == CONNECTION_ERROR);
}

void handleConectionErrors() {
  if (msgState == CONNECTION_ERROR) {
#ifdef DEBUG_RADIO
    Serial.println("handleConectionErrors CONNECTION_ERROR");
#endif
    if (!retryTimer.isRunning()) { // have not started retry wait yet
      retryCounter++;
      if (retryCounter < MAX_RETRIES) {
#ifdef DEBUG_RADIO
        Serial.println("retry");
#endif
        msgState = RETRY_WAIT;
        setFlash(SLOW_FLASH);
        setBuzzer(BUZZER_MSG_RETRY);
        // try to connect again after small wait (1sec)
        retryTimer.start(RETRY_TIMER_INTERVAL);
        offDelay.start(RETRY_TIMER_INTERVAL + MSG_TIMEOUT + 1000); // stary on for at least this time
        return;
      } // else stop now
    }
    // else stop after flashing for 20sec
    disableTimers();
    offDelay.start(ERROR_OFF_TIMER_INTERVAL);
    setFlash(FAST_FLASH);
    setBuzzer(BUZZER_ERROR);
#ifdef DEBUG_RADIO
    Serial.println("   setBuzzer(BUZZER_ERROR)");
#endif
    msgState = CONNECTION_SEND_CLOSE;
  }
}

void disableRadioTimers() {
  travelTimer.stop();
  retryTimer.stop();
}

void checkRetryTimer() {
  if (msgState == RETRY_WAIT) {
    if (retryTimer.justFinished()) {
      disableTimers();
      // try to connect again
#ifdef DEBUG_RADIO
      Serial.println("Retry Timed out set CONNECTED again");
#endif
      msgState = CONNECTED;
    }
  }
}

void clearClientResponse() {
  parseReturnInit(); // clear out parser
  while (pfodClient.available()) {
    pfodClient.read();
  }
}

//https://stackoverflow.com/questions/4836534/returning-a-pointer-to-a-literal-or-constant-character-array-string
//String literals are guaranteed to live for the span of the application. It might not be in read-only/immutable memory,
//but they're guaranteed to exist for the lifetime of the application, so returning a pointer to one is fine
//CONNECTED, SENT_CMD, CONNECTION_FINISHED, CONNECTION_ERROR, CONNECTION_SEND_CLOSE, CHARGING
const char* debugMsgState() {
  switch (msgState) {
    case CONNECTED:
      return "CONNECTED";

    case SENT_INQUIRY:
      return "SENT_INQUIRY";

    case SENT_CMD:
      return "SENT_CMD";

    case RETRY_WAIT:
      return "RETRY_WAIT";

    case CONNECTION_FINISHED:
      return "CONNECTION_FINISHED";

    case CONNECTION_ERROR:
      return "CONNECTION_ERROR";

    case CONNECTION_SEND_CLOSE:
      return "CONNECTION_SEND_CLOSE";

    case CHARGING:
      return "CHARGING";

    default:
      return "_UNKNOWN_";
  }
}

void sendCloseConnectionOnError() {
  if (msgState == CONNECTION_SEND_CLOSE) {
    pfodClient.print(CLOSE_CONNECTION_MSG);
#ifdef DEBUG_RADIO
    Serial.println(CLOSE_CONNECTION_MSG);
#endif
    msgState = CONNECTION_FINISHED;
  }
}



void reachedRequestedPosition() {
  msgState = CONNECTION_SEND_CLOSE; //
  lastRequestedDoorState = ' ';
  requestedDoorState = ' ';
  disableTimers();
  offDelay.start(MIN_OFF_TIMER_INTERVAL);
}

void processMsgs() {
  uint8_t cmd  = 0;

  if ((msgState == CONNECTION_ERROR) || (msgState == CONNECTION_FINISHED) || (msgState == CONNECTION_SEND_CLOSE) || (msgState == RETRY_WAIT) || (msgState == CHARGING) ) {
    return;
  }

  if (requestedDoorState == ' ') {
    // nothing to do
  }

  if (msgState == CONNECTED) { // always do this once on start
    // clear input
    clearClientResponse();
    bool USB_cmd = false;
#ifdef DEBUG_RADIO
    if (Serial.available()) {
      Serial.println();
      Serial.print("Sending: ");
      while (Serial.available()) {
        uint8_t c = Serial.read();
        pfodClient.write((uint8_t)c);
        Serial.write((char)c);
      }
      USB_cmd = true; // skip dwg update
    }
#endif
    if (!USB_cmd) {
      pfodClient.print(REQUEST_IMAGE_UPDATE_MSG); // ask for image z update for version V3
#ifdef DEBUG_RADIO
      Serial.println(REQUEST_IMAGE_UPDATE_MSG);
#endif
    }
    msgState = SENT_INQUIRY;
    msgTimer.start(MSG_TIMEOUT);
#ifdef DEBUG_RADIO
    Serial.print("msgTimer:"); Serial.println(msgTimer.remaining());
#endif
    return;
  }

  if ((msgState == SENT_CMD) || (msgState == SENT_INQUIRY)) {
    if (msgTimer.justFinished()) {
#ifdef DEBUG_RADIO
      Serial.println("SENT_CMD/SENT_INQUIRY return timed out.");
#endif
      msgState = CONNECTION_ERROR; // did not get return msg
      // for testing bad connections
      //      if (retryCounter > 5) {
      //        radio.connectTo( 0x6b); // set the server address to connect to
      //      }
      return;
    }
    while (pfodClient.available() && (cmd == 0)) {
      char c = pfodClient.read();
#ifdef DEBUG_RADIO
      Serial.print(c);
#endif
      cmd = parseReturn(c);
    }
    if (cmd != 0) {
      // have whole msg
      msgTimer.stop(); // got response to msg

      msgStateEnum lastMsgState = msgState;
      msgState = CONNECTED;

      setFlash(NO_FLASH); // solid if have connection

      if (cmd != '+') {
#ifdef DEBUG_RADIO
        Serial.println(); Serial.println("SENT_CMD return was not {+ .. ");
#endif
        msgState = CONNECTION_ERROR; // did not get {+ msg back
        return;
      }

      doorState = checkDoorState(); // look for onOff state
      if (doorState == _UNKNOWN_) {
#ifdef DEBUG_RADIO
        Serial.println("SENT_CMD return did not contain door state");
#endif
        msgState = CONNECTION_ERROR; // did not find ligth state in return msg
        return;
      }

      //  test code for cmd not responded to and start retries.
      //      if (lastMsgState == SENT_CMD) {
      //        if (retryCounter < 5) {
      //          msgState = CONNECTION_ERROR;
      //          return;
      //        }
      //      }

      // else got valid response so if was SEND_CMD clear requestedDoorState
      if (lastMsgState == SENT_CMD) {
        requestedDoorState = ' ';
        disableTimers();
        // start travel timer now that cmd has been responded to
        // if send another cmd this will reset again when response to that one received
        travelTimer.start(TRAVEL_TIMER_INTERVAL); // allow time for cmd to work
#ifdef DEBUG_RADIO
        Serial.println("Got replay to Cmd");
        printDebugTimers();
#endif
      }

      // check if have reached correct state
      // this could happen with no cmd being sent.
      if (lastRequestedDoorState == 'O') { // opened
        if (doorState == OPEN) {
#ifdef DEBUG_RADIO
          Serial.println("Door is in correct state OPEN");
#endif
          reachedRequestedPosition();
        }
      } else if (lastRequestedDoorState == 'C') {
        if (doorState == CLOSED) {
#ifdef DEBUG_RADIO
          Serial.println("Door is in correct state CLOSED");
#endif
          reachedRequestedPosition();
        }
      }

      if (requestedDoorState != ' ') {
        // send request
        if (requestedDoorState == 'C') {
          pfodClient.print(CLOSE_DOOR);
#ifdef DEBUG_RADIO
          Serial.println(CLOSE_DOOR);
#endif
        } else if (requestedDoorState == 'O') {
          pfodClient.print(OPEN_DOOR);
#ifdef DEBUG_RADIO
          Serial.println(OPEN_DOOR);
#endif
        }
        msgState = SENT_CMD;
        // don't clear requested state until get response to SENT_CMD
        // allows for retries if connection failes
        // another button press will resend
#ifdef DEBUG_RADIO
        Serial.println("Send cmd");
        printDebugTimers();
#endif

      }
    }
    // else ignore
    return;
  }
  // other msg state ignore
}




enum parserStateEnum {LOOKING_FOR_START, IN_MSG, PARSER_FINISHED, PARSER_ERROR};
parserStateEnum parserState = LOOKING_FOR_START;
uint8_t pfodMsgStarted = '{';
uint8_t pfodMsgEnd = '}';
size_t returnMsgIdx = 0;

void parseReturnInit() {
  returnMsgIdx = 0;
  returnMsg[0] = '\0';
  parserState = LOOKING_FOR_START;
}


/**
   Look for { ... } msg, max 255, and return first char after {
   save whole msg in buffer (max 255)
*/
uint8_t parseReturn(uint8_t in) {
  if ((parserState == LOOKING_FOR_START) || (parserState == PARSER_FINISHED) ) {
    parserState = LOOKING_FOR_START;
    if (in == pfodMsgStarted) { // found {
      parseReturnInit(); // clean out last cmd
      parserState = IN_MSG;
      returnMsg[returnMsgIdx++] = in; // save {
    }
    // else ignore this char as waiting for start {
    return 0;
  }
  if (parserState == IN_MSG) {
    // else have seen {
    if ((returnMsgIdx >= (MAX_RETURN_MSG_LEN - 1)) && (in != pfodMsgEnd)) {
      // ignore this msg and reset
      returnMsg[returnMsgIdx++] = 0; // termimate
      parserState = PARSER_ERROR;
#ifdef DEBUG_RADIO
      Serial.print("Return msg not completed after "); Serial.print(MAX_RETURN_MSG_LEN); Serial.print(" bytes.");
#endif
    } else {
      returnMsg[returnMsgIdx++] = in;
    }
    // look for char }
    if (in == pfodMsgEnd) {
      returnMsg[returnMsgIdx++] = 0; // termimate
      parserState = PARSER_FINISHED; // will reset parser on next call
    }
  }

  // now choose return based on current state
  if (parserState == PARSER_FINISHED) {
#ifdef DEBUG_RADIO
    Serial.print("parser returned '");   Serial.print((char)returnMsg[1]);   Serial.println('\'');
#endif
    return returnMsg[1]; // first char after {
  } else if (parserState == PARSER_ERROR) {
    return 0xff; // error return paser will not be rest next call will need to call returnParserInit()
  } else {
    return 0; // no complete msg yet
  }
}



/**
   returns 0 for Closed, 1 for Open and -1 for error (not found)
*/
DoorStateEnum checkDoorState() {
#ifdef DEBUG_RADIO
  Serial.print("Door is ");
#endif
  char* found = NULL;
  // check for ON
  found = strstr(returnMsg, DOOR_CLOSED );
  if (found != NULL) {
#ifdef DEBUG_RADIO
    Serial.println(" Closed");
#endif
    return CLOSED;
  }
  found = strstr(returnMsg, DOOR_OPEN );
  if (found != NULL) {
#ifdef DEBUG_RADIO
    Serial.println("Open");
#endif
    return OPEN;
  }
  found = strstr(returnMsg, DOOR_OPENING);
  if (found != NULL) {
#ifdef DEBUG_RADIO
    Serial.println("Opening");
#endif
    return OPENING;
  }
  found = strstr(returnMsg, DOOR_CLOSING);
  if (found != NULL) {
#ifdef DEBUG_RADIO
    Serial.println("Closing");
#endif
    return CLOSING;
  }
  found = strstr(returnMsg, DOOR_STOPPED);
  if (found != NULL) {
#ifdef DEBUG_RADIO
    Serial.println("Stopped");
#endif
    return STOPPED;
  }
  // else not found
  return _UNKNOWN_;
}


void checkTravelTime() {
  if (travelTimer.justFinished()) {
    disableTimers();
#ifdef DEBUG_RADIO
    Serial.println("Travel Time timed out");
#endif
    offDelay.start(ERROR_OFF_TIMER_INTERVAL);
    setFlash(FAST_FLASH);
    setBuzzer(BUZZER_ERROR);
#ifdef DEBUG_RADIO
    Serial.println("   setBuzzer(BUZZER_ERROR)");
#endif
    msgState = CONNECTION_SEND_CLOSE;
  }
}


void checkRadio() {
  checkRetryTimer();
  processMsgs();
  handleConectionErrors();
  sendCloseConnectionOnError(); //if set from processMsgs();
  checkTravelTime(); // has the door reached the required position in this time

  if (rawDataStream->available()) {
    while (rawDataStream->available()) {
      int b = rawDataStream->read();
      //Serial.print((char)b);  // comment this line out if you want to ignore any RawData sent by pfodServer
    }
  }
}

// implementation of pfodRHDriver
pfodRHDriver::pfodRHDriver(RHGenericDriver * _driver) {
  driver = _driver;
}
bool pfodRHDriver::init() {
  return driver->init();
}
int16_t pfodRHDriver::lastRssi() {
  return driver->lastRssi();
}
int pfodRHDriver::getMode() {
  return driver->mode();
}
uint8_t pfodRHDriver::getMaxMessageLength() {
  return driver->maxMessageLength();
}
void pfodRHDriver::setThisAddress(uint8_t addr) {
  driver->setThisAddress(addr);
}
bool pfodRHDriver::receive(uint8_t* buf, uint8_t* len) {
  return driver->recv(buf, len);
}
bool pfodRHDriver::send(const uint8_t* data, uint8_t len) {
  return driver->send(data, len);
}
uint8_t pfodRHDriver::headerTo() {
  return driver->headerTo();
}
uint8_t pfodRHDriver::headerFrom() {
  return driver->headerFrom();
}
uint8_t pfodRHDriver::headerId() {
  return driver->headerId();
}
uint8_t pfodRHDriver::headerFlags() {
  return driver->headerFlags();
}
void pfodRHDriver::setHeaderTo(uint8_t to) {
  driver->setHeaderTo(to);
}
void pfodRHDriver::setHeaderFrom(uint8_t from) {
  driver->setHeaderFrom(from);
}
void pfodRHDriver::setHeaderId(uint8_t id) {
  driver->setHeaderId(id);
}
void pfodRHDriver::setHeaderFlags(uint8_t flags) {
  driver->setHeaderFlags(flags, 0xff);
}
// end of pfodRHDriver implementation
